from collections import OrderedDict
import os.path

from .compat import (
    urlencode,
    urlparse,
)
from .exceptions import InvalidURI


class PlasterURL(object):
    """
    Represents the components of a URL used to locate a
    :class:`plaster.ILoader`.

    :ivar scheme: The name of the loader backend.

    :ivar path: The loader-specific path string.
        This is the entirety of the ``config_uri`` passed to
        :func:`plaster.parse_uri` without the scheme, fragment and options.
        If this value is falsey it is replaced with an empty string.

    :ivar options: A dictionary of options parsed from the query string as
        url-encoded key=value pairs.

    :ivar fragment: A loader-specific default section name.
        This parameter may be used by loaders in scenarios where they provide
        APIs that support a default name. For example, a loader that provides
        ``get_wsgi_app`` may use the fragment to determine the name of the
        section containing the WSGI app if none was explicitly defined.
        If this value is falsey it is replaced with an empty string.

    """

    def __init__(self, scheme, path='', options=None, fragment=''):
        self.scheme = scheme
        if not path:
            path = ''
        self.path = path
        if options is None:
            options = {}
        self.options = options
        if not fragment:
            fragment = ''
        self.fragment = fragment

    def __str__(self):
        result = '{0.scheme}://{0.path}'.format(self)
        if self.options:
            result += '?' + urlencode(self.options)
        if self.fragment:
            result += '#' + self.fragment
        return result

    def __repr__(self):
        return 'PlasterURL(\'{0}\')'.format(self)


def parse_uri(config_uri):
    """
    Parse the ``config_uri`` into a :class:`plaster.PlasterURL` object.

    ``config_uri`` can be a relative or absolute file path such as
    ``development.ini`` or ``/path/to/development.ini``. The file must have
    an extension that can be handled by a :class:`plaster.ILoader`
    registered with the system.

    Alternatively, ``config_uri`` may be a :rfc:`1738`-style string.

    """
    if isinstance(config_uri, PlasterURL):
        return config_uri

    # force absolute paths to look like a uri for more accurate parsing
    # we throw away the dummy scheme later and parse it from the resolved
    # path extension
    isabs = os.path.isabs(config_uri)
    if isabs:
        config_uri = 'dummy://' + config_uri

    # check if the uri is actually a url
    parts = urlparse.urlparse(config_uri)

    # reconstruct the path without the scheme and fragment
    path = urlparse.ParseResult(
        scheme='',
        netloc=parts.netloc,
        path=parts.path,
        params='',
        query='',
        fragment='',
    ).geturl()
    # strip off leading //
    if path.startswith('//'):
        path = path[2:]

    if parts.scheme and not isabs:
        scheme = parts.scheme

    else:
        scheme = os.path.splitext(path)[1]
        if scheme.startswith('.'):
            scheme = scheme[1:]

        # tag uris coming from file extension as file+scheme
        if scheme:
            scheme = 'file+' + scheme

    query = parts.query if parts.query else None
    options = OrderedDict()
    if query:
        options.update(urlparse.parse_qsl(query))
    fragment = parts.fragment if parts.fragment else None

    if not scheme:
        raise InvalidURI(config_uri, (
            'Could not determine the loader scheme for the supplied '
            'config_uri "{0}"'.format(config_uri)))

    return PlasterURL(
        scheme=scheme,
        path=path,
        options=options,
        fragment=fragment,
    )



# flake8: noqa
import sys

PY2 = sys.version_info[0] == 2


def add_metaclass(metaclass):  # pragma: no cover
    """Class decorator for creating a class with a metaclass."""

    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)

    return wrapper


if PY2:
    import urlparse
    from urllib import urlencode
else:
    from urllib import parse
    urlparse = parse
    urlencode = parse.urlencode



import abc

from .compat import add_metaclass


@add_metaclass(abc.ABCMeta)
class ILoader(object):
    """
    An abstraction over an source of configuration settings.

    It is required to implement ``get_sections``, ``get_settings`` and
    ``setup_logging``.

    Optionally, it may also implement other :term:`loader protocol` interfaces
    to provide extra functionality. For example,
    :class:`plaster.protocols.IWSGIProtocol` which requires ``get_wsgi_app``,
    and ``get_wsgi_server`` for loading WSGI configurations. Services that
    depend on such functionality should document the required functionality
    behind a particular :term:`loader protocol` which custom loaders can
    implement.

    :ivar uri: The :class:`plaster.PlasterURL` object used to find the
        :class:`plaster.ILoaderFactory`.

    """

    @abc.abstractmethod
    def get_sections(self):
        """
        Load the list of section names available.

        """

    @abc.abstractmethod
    def get_settings(self, section=None, defaults=None):
        """
        Load the settings for the named ``section``.

        :param section: The name of the section in the config file. If this is
            ``None`` then it is up to the loader to determine a sensible
            default usually derived from the fragment in the ``path#name``
            syntax of the ``config_uri``.

        :param defaults: A ``dict`` of default values used to populate the
            settings and support variable interpolation. Any values in
            ``defaults`` may be overridden by the loader prior to returning
            the final configuration dictionary.

        :returns: A ``dict`` of settings. This should return a dictionary
            object even if the section is missing.
        :raises ValueError: If a section name is missing and cannot be
            determined from the ``config_uri``.

        """

    @abc.abstractmethod
    def setup_logging(self, defaults=None):
        """
        Execute the logging configuration defined in the config file.

        This function should, at least, configure the Python standard logging
        module. However, it may also be used to configure any other logging
        subsystems that serve a similar purpose.

        :param defaults: A ``dict`` of default values used to populate the
            settings and support variable interpolation. Any values in
            ``defaults`` may be overridden by the loader prior to returning
            the final configuration dictionary.

        """


@add_metaclass(abc.ABCMeta)
class ILoaderFactory(object):
    @abc.abstractmethod
    def __call__(self, uri):
        """
        A factory which accepts a :class:`plaster.PlasterURL` and returns a
        :class:`plaster.ILoader` object.

        """


@add_metaclass(abc.ABCMeta)
class ILoaderInfo(object):
    """
    An info object describing a specific :class:`plaster.ILoader`.

    :ivar scheme: The full scheme of the loader.
    :ivar protocols: Zero or more supported :term:`loader protocol`
        identifiers.
    :ivar factory: The :class:`plaster.ILoaderFactory`.

    """

    @abc.abstractmethod
    def load(self, config_uri):
        """
        Create and return an :class:`plaster.ILoader` instance.

        :param config_uri: Anything that can be parsed by
            :func:`plaster.parse_uri`.

        """
